/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComScale.cpp
    \brief    Functions for upscale and downscale of pictures for dynamic resolution coding
*/

#include <TComScale.h>
#include <assert.h>


long double upScaleComponent2x2( const Pel* pSrc, Pel* pDest, const Int iSrcStride, const Int iDestStride, const Int iSrcWidth, const Int iSrcHeight){
  const int iDestHeight = iSrcHeight*2;

  const int NUM_UP_TAPS = 2;
  const int iTaps[NUM_UP_TAPS] = {19,-3};
  const int iFilterShift = 5;

  const int iTempStride = iSrcWidth+2*NUM_UP_TAPS;
  Pel* pTemp = new Pel[ iDestHeight * iTempStride];
  // Vertical
  const Pel* pDataSrc = pSrc;
  Pel* pDataDest = pTemp + NUM_UP_TAPS;
  int iTemp;

  for (int j=0; j<iSrcHeight; ++j){
    // Zero phase
    for (int i=0; i<iSrcWidth; ++i){
      pDataDest[i] = pDataSrc[i];
    }
    std::fill(pDataDest - NUM_UP_TAPS, pDataDest, pDataDest[0] );
    std::fill(pDataDest + iSrcWidth, pDataDest+iSrcWidth+NUM_UP_TAPS, pDataDest[iSrcWidth-1] );
    pDataDest += iTempStride;

    // +1/2 phase
    for (int i=0; i<iSrcWidth; ++i){
      iTemp = 0;
      for (int k=0; k<NUM_UP_TAPS; ++k){
        iTemp += iTaps[k] * (pDataSrc[i + (k+1)*iSrcStride]  + pDataSrc[i - k*iSrcStride] );
      }
      iTemp += 1<<(iFilterShift-1);
      pDataDest[i] = Clip(iTemp>>iFilterShift);
    }
    std::fill(pDataDest - NUM_UP_TAPS, pDataDest, pDataDest[0] );
    std::fill(pDataDest + iSrcWidth, pDataDest+iSrcWidth+NUM_UP_TAPS, pDataDest[iSrcWidth-1] );

    pDataDest += iTempStride;
    pDataSrc += iSrcStride;
  }

  pDataSrc = pTemp + NUM_UP_TAPS;
  pDataDest = pDest;

  for (int j=0; j<iDestHeight; ++j){
    for (int i=0,ii=0; i<iSrcWidth; ++i, ii+=2){
      pDataDest[ii] = pDataSrc[i];
      iTemp = 0;
      for (int k=0; k<NUM_UP_TAPS; ++k){
        iTemp += iTaps[k] * (pDataSrc[i - k] + pDataSrc[i + k+1] );
      }
      iTemp += 1<<(iFilterShift-1);
      pDataDest[ii+1] = Clip(iTemp>>iFilterShift);

    }
    pDataSrc += iTempStride;
    pDataDest += iDestStride;
  }

  delete[] pTemp;

}

void upScale(const Int iSrcResIdx, const Int iDestResIdx, TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDest){

  if (iSrcResIdx==1 && iDestResIdx == 0){// FIXME - make more general
    pcYuvSrc->extendPicBorder();
    upScaleComponent2x2( pcYuvSrc->getLumaAddr(), pcYuvDest->getLumaAddr(), pcYuvSrc->getStride(), pcYuvDest->getStride(), pcYuvSrc->getWidth(), pcYuvSrc->getHeight() );
    upScaleComponent2x2( pcYuvSrc->getCbAddr(), pcYuvDest->getCbAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 );
    upScaleComponent2x2( pcYuvSrc->getCrAddr(), pcYuvDest->getCrAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 );
    pcYuvDest->extendPicBorder();
  }
}

void downScaleComponent2x2( const Pel* pSrc, Pel* pDest, const Int iSrcStride, const Int iDestStride, const Int iSrcWidth, const Int iSrcHeight){

  const int iDestWidth = iSrcWidth/2;
  const int iDestHeight = iSrcHeight/2;

//  const int HALF_NUM_DOWN_TAPS = 6;
//  const int iTaps[HALF_NUM_DOWN_TAPS] = {19, 5, -3, -4, 0 , 2};
//  const int iCentreTap = 26;
//  const int iFilterShift = 6;

  const int HALF_NUM_DOWN_TAPS = 2;
  const int iTaps[HALF_NUM_DOWN_TAPS] = {9, -1};
  const int iCentreTap = 16;
  const int iFilterShift = 5;

  const int iTempStride = iSrcWidth + 2*HALF_NUM_DOWN_TAPS;
  Pel* pTemp = new Pel[ iDestHeight * iTempStride];

  int iTemp;

  // Vertical
  const Pel* pDataSrc = pSrc;
  Pel* pDataDest = pTemp + HALF_NUM_DOWN_TAPS;

  for (int j=0; j<iDestHeight; ++j){
    for (int i=0; i<iSrcWidth; ++i){
      iTemp = pDataSrc[i]*iCentreTap;
      for (int k=0; k<HALF_NUM_DOWN_TAPS; ++k){
        iTemp += iTaps[k]*(pDataSrc[i+(k+1)*iSrcStride] + pDataSrc[i-(k+1)*iSrcStride]);
      }
      iTemp += 1<<(iFilterShift-1);
      pDataDest[i] = Clip(iTemp>>iFilterShift);
    }
    std::fill(pDataDest - HALF_NUM_DOWN_TAPS, pDataDest, pDataDest[0] );
    std::fill(pDataDest + iSrcWidth, pDataDest+iSrcWidth+HALF_NUM_DOWN_TAPS, pDataDest[iSrcWidth-1] );
    pDataSrc += 2*iSrcStride;
    pDataDest += iTempStride;
  }

  // Horizontal
  pDataSrc = pTemp+HALF_NUM_DOWN_TAPS;
  pDataDest = pDest;

  for (int j=0; j<iDestHeight; ++j){
    for (int i=0,ii=0; i<iDestWidth; ++i,ii+=2){
      iTemp = 0;
      iTemp = pDataSrc[ii]*iCentreTap;
      for (int k=0; k<HALF_NUM_DOWN_TAPS; ++k){
        iTemp += iTaps[k]*(pDataSrc[ii+k+1] + pDataSrc[ii-k-1]);
      }
      iTemp += 1<<(iFilterShift-1);
      pDataDest[i] = Clip(iTemp>>iFilterShift);

    }
    pDataSrc += iTempStride;
    pDataDest += iDestStride;
  }

  delete[] pTemp;

}

void downScale(const Int iSrcResIdx, const Int iDestResIdx, TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDest){

  if (iSrcResIdx==0 && iDestResIdx == 1){// FIXME - make more general
    pcYuvSrc->extendPicBorder();
    downScaleComponent2x2( pcYuvSrc->getLumaAddr(), pcYuvDest->getLumaAddr(), pcYuvSrc->getStride(), pcYuvDest->getStride(), pcYuvSrc->getWidth(), pcYuvSrc->getHeight() );
    downScaleComponent2x2( pcYuvSrc->getCbAddr(), pcYuvDest->getCbAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 );
    downScaleComponent2x2( pcYuvSrc->getCrAddr(), pcYuvDest->getCrAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 );
    pcYuvDest->extendPicBorder();
  }
}

